/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 *  Gabber 2
 *  Based on Gabber, Copyright (c) 1999-2002 Dave Smith & Julian Missig
 *  Copyright (c) 2002-2003 Julian Missig
 */

#include "ContactInfo.hh"

#include "JabberConnection.hh"

#include "GabberUtility.hh"
#include "GabberWidgets.hh"

#include <jabberoo/jabberoo.hh>
#include <jabberoo/vCard.h>

#include <gtkmm/dialog.h>
#include <gtkmm/sizegroup.h>
#include <gtkmm/textview.h>
#include <gtkmm/window.h>
#include <libintl.h>

using namespace jabberoo;

namespace Gabber {

// ---------------------------------------------------------
//
// Contact Information Dialog
//
// ---------------------------------------------------------

void ContactInfoDlg::display(JabberConnection& conn, Gtk::Window& parent,
			     const Glib::ustring& jid)
{
     ContactInfoDlg* e = manage(new ContactInfoDlg(conn, parent, jid));
}

ContactInfoDlg::ContactInfoDlg(JabberConnection& conn, Gtk::Window& parent,
			       const Glib::ustring& jid)
     : BaseGabberWindow("ViewInfo_dlg"), _conn(conn), _jid(jid)
{
     // Set the dialog's parent
     Gtk::Dialog* dlg  = static_cast<Gtk::Dialog*>(getGtkWindow());
     dlg->set_transient_for(parent);

     Gtk::Button* b;
     get_widget("Close_btn", b);
     b->signal_clicked().connect(slot(*this, &ContactInfoDlg::close));

     // Align the labels with the magic of SizeGroup
     Gtk::Label* l;
     Glib::RefPtr<Gtk::SizeGroup> client_grp = Gtk::SizeGroup::create(Gtk::SIZE_GROUP_HORIZONTAL);
     get_widget("JabberClientName_lbl_lbl", l);
     client_grp->add_widget(*l);
     get_widget("JabberClientVersion_lbl_lbl", l);
     client_grp->add_widget(*l);
     get_widget("ComputerOS_lbl_lbl");
     client_grp->add_widget(*l);
     get_widget("ComputerTime_lbl_lbl");
     client_grp->add_widget(*l);

     _prettyjid = manage(new PrettyJID(_jid, "", PrettyJID::dtJIDRes, true));
     _prettyjid->show();
     _prettyjid->changed.connect(slot(*this, &ContactInfoDlg::on_PrettyJID_changed));
     on_PrettyJID_changed(); // Fire once since this won't normally be fired when we init

     Gtk::HBox* hb;
     get_widget("JIDInfo_hbox", hb);
     hb->pack_start(*_prettyjid, true, true, 0);
     _nickname = _prettyjid->get_nickname();

     // Set the window title
     _thisWindow->set_title(Util::substitute(_("%s's Info"), _nickname));

     Gtk::Entry* nickent;
     get_widget("Nickname_ent", nickent);
     nickent->set_text(_nickname);

     send_vcard_request();
}

void ContactInfoDlg::send_vcard_request()
{
     // Get next session ID
     string id = _conn.getSession().getNextID();

     // Construct vCard request
     Packet iq("iq");
     iq.setID(id);
     iq.setTo(_jid);
     iq.getBaseElement().putAttrib("type", "get");
     Element* vCard = iq.getBaseElement().addElement("vCard");
     vCard->putAttrib("xmlns", "vcard-temp");
     vCard->putAttrib("version", "2.0");
     vCard->putAttrib("prodid", "-//HandGen//NONSGML vGen v1.0//EN");

     // Send the vCard request
     _conn.getSession() << iq;
     _conn.getSession().registerIQ(id, slot(*this, &ContactInfoDlg::parse_vcard));
}

void ContactInfoDlg::set_info_label(const Glib::ustring& label_name, const Glib::ustring& data)
{
     if (data.empty())
	  return;

     Gtk::Label* l;
     get_widget(label_name + "_lbl", l);
     l->show();
     get_widget(label_name, l);
     l->set_label(data);
     l->show();
}

void ContactInfoDlg::parse_vcard(const judo::Element& t)
{
     if (t.empty())
	  return;

     const judo::Element* vCardElem = t.findElement("vCard");
     if (vCardElem == NULL)
	  return;

     Glib::ustring curtext;
 
     vCard vc(*vCardElem);

     // General
     curtext = vc[vCard::Fullname];
     set_info_label("FullName_lbl", curtext);
     curtext = vc[vCard::EMail];
     set_info_label("Email_lbl", curtext);
     curtext = vc[vCard::URL];
     set_info_label("WebSite_lbl", curtext);
     curtext = vc[vCard::Telephone];
     set_info_label("Telephone_lbl", curtext);

     // About
     curtext = vc[vCard::Birthday];
     if (!curtext.empty())
     {
	  Gtk::Widget* f;
	  get_widget("Birthday_frame", f);
	  f->show();
	  set_info_label("Birthday_lbl", curtext);
     }
     curtext = vc[vCard::Description];
     Gtk::TextView* tv;
     get_widget("About_txtview", tv);
     tv->get_buffer()->set_text(curtext);
}

void ContactInfoDlg::send_version_request()
{
     // Get next session ID
     string id = _conn.getSession().getNextID();

     // Construct version request
     Packet iq("iq");
     iq.setID(id);
     iq.setTo(_prettyjid->get_full_jid());
     iq.getBaseElement().putAttrib("type", "get");
     Element* vCard = iq.getBaseElement().addElement("query");
     vCard->putAttrib("xmlns", "jabber:iq:version");

     // Send the version request
     _conn.getSession() << iq;
     _conn.getSession().registerIQ(id, slot(*this, &ContactInfoDlg::parse_version));
}

void ContactInfoDlg::parse_version(const judo::Element& t)
{
     if (t.empty())
	  return;

     const Element* query = t.findElement("query");
     if (query == NULL)
	  return;
	 
     set_info_label("JabberClientName_lbl", query->getChildCData("name"));
     set_info_label("JabberClientVersion_lbl", query->getChildCData("version"));
     set_info_label("ComputerOS_lbl", query->getChildCData("os"));
}

void ContactInfoDlg::send_time_request()
{
     // Get next session ID
     string id = _conn.getSession().getNextID();

     // Construct version request
     Packet iq("iq");
     iq.setID(id);
     iq.setTo(_prettyjid->get_full_jid());
     iq.getBaseElement().putAttrib("type", "get");
     Element* vCard = iq.getBaseElement().addElement("query");
     vCard->putAttrib("xmlns", "jabber:iq:time");

     // Send the version request
     _conn.getSession() << iq;
     _conn.getSession().registerIQ(id, slot(*this, &ContactInfoDlg::parse_time));
}

void ContactInfoDlg::parse_time(const judo::Element& t)
{
     if (t.empty())
	  return;

     const Element* query = t.findElement("query");
     if (query == NULL)
	  return;

     Glib::ustring curtext;
     curtext = query->getChildCData("display");
     curtext += " "; // Don't duplicate the time zone:
     if(curtext.find(query->getChildCData("tz"))==string::npos)
	  curtext += query->getChildCData("tz");
     set_info_label("ComputerTime_lbl", curtext);
}

void ContactInfoDlg::on_PrettyJID_changed()
{
     send_version_request();
//     send_last_request();
     send_time_request();
//     get_status();
}

} // namespace Gabber
