/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 *  Gabber 2
 *  Based on Gabber, Copyright (c) 1999-2002 Dave Smith & Julian Missig
 *  Copyright (c) 2002 Julian Missig
 */

#include "JabberConnection.hh"
#include "Environment.hh"
#include "PrefsInterface.hh"
#include "ConfigPaths.hh"
#include "ConfigWrapper.hh"

#include <sys/utsname.h>

namespace Gabber {

// -------------------------------------------------------------------
//
// Generic Jabber Connection = Jabberoo + TCPTransmitter
//
// -------------------------------------------------------------------
JabberConnection::JabberConnection() : ConfigWrapper(Keys::acct.main_dir)
{
     // Instiatiate objects
     _Session     = new jabberoo::Session;
     _Transmitter = new TCPTransmitter;

     // Debugging signals
     _Session->evtXMLParserError.connect(slot(*this, &JabberConnection::on_XML_ParserError));

     // Transmitter signals
     _Transmitter->evtConnected.connect(slot(*this, &JabberConnection::on_transmitter_connected));
     _Transmitter->evtDisconnected.connect(slot(*this, &JabberConnection::on_transmitter_disconnected));
     _Transmitter->evtReconnect.connect(slot(*this, &JabberConnection::on_transmitter_reconnect));
     _Transmitter->evtError.connect(slot(*this, &JabberConnection::on_transmitter_error));
     _Transmitter->evtDataAvailable.connect(slot(*_Session, &jabberoo::Session::push));

     // Session signals
     _Session->evtTransmitXML.connect(slot(*_Transmitter, &TCPTransmitter::send));
     _Session->evtConnected.connect(slot(*this, &JabberConnection::on_session_connected));
     _Session->evtDisconnected.connect(slot(*this, &JabberConnection::on_session_disconnected));
     _Session->evtOnVersion.connect(slot(*this, &JabberConnection::on_session_version));

     _Session->evtOnTime.connect(slot(*this, &JabberConnection::on_session_time));
}

JabberConnection::~JabberConnection()
{
     // <aolvoice>Goodbye</aolvoice>
     delete _Session;
     _Session = NULL; // Just in case delete _Transmitter sends a signal
     delete _Transmitter;
}

void JabberConnection::connect(bool login)
{
     // Whether or not we want to attempt to log in
     _login = login;

     // set the proxy type
     // _Transmitter->setProxy...

     // tell stuff we're connecting to the server
     evtConnecting();

     // Connect the transmitter
     _Transmitter->connect(client->get_string(Keys::acct.server),
			   client->get_int(Keys::acct.port),
			   false,
			   false);
     // Start polling to attempt to keep the connection open
     _Transmitter->startPoll();

     // JabberConnection::on_transmitter_connected is next
}

// -------------------------------------
// Debugging event handlers
// -------------------------------------
void JabberConnection::on_XML_ParserError(int error_code, const string& error_msg)
{
     // Display an error explaining that there was an error parsing the XML
     // TODO: see above ;)

     // Try to disconnect as gracefully as possible
     // TODO: set presence to offline
     if (!_Session->disconnect())
	  _Transmitter->disconnect();
}


// -------------------------------------
// Transmitter event handlers
// -------------------------------------
void JabberConnection::on_transmitter_connected()
{
     // check for the password
     if (client->get_string(Keys::acct.password).empty())
     {
	  // ask for the password
	  LoginDlg::execute();
     }
     // Figure out the auth method
     // Default to auto-select: 0k, digest, plaintext
     jabberoo::Session::AuthType atype = jabberoo::Session::atAutoAuth;
     // if they want to force one type, do so

     // Tell them we're logging in
     evtLoggingIn();
     // Send the credentials and attempt to log in
     _Session->connect(client->get_string(Keys::acct.server),
		       atype,
		       client->get_string(Keys::acct.username),
		       client->get_string(Keys::acct.resource),
		       client->get_string(Keys::acct.password),
		       false,
		       _login);
}

void JabberConnection::on_transmitter_disconnected()
{
     // only do this if the session exists
     if (_Session)
     {
          // Tell the session to disconnect
          cout << "DISCONNECTED" << endl;
          _Session->disconnect();
     }
     
     // Tell the user we've been disconnected
     // this is *not* thrown for on_session_disconnected because
     // that event will end up causing this one anyway
     evtDisconnected();
}

void JabberConnection::on_transmitter_reconnect()
{
     // Tell the user we're reconnecting
     evtReconnecting();
}

void JabberConnection::on_transmitter_error(const string & emsg)
{
     cout << "Transmitter error. Disconnected: " << emsg << endl;
     _Transmitter->disconnect();
}

// -------------------------------------
// Session event handlers
// -------------------------------------
void JabberConnection::on_session_connected(const judo::Element& t)
{
     _connected = true;
     // We are technically connected at this point
     // however, we really don't want things to go on
     // until /after/ the roster has been loaded
     _roster_conn = _Session->evtOnRoster.connect(slot(*this, &JabberConnection::on_session_roster));
}

void JabberConnection::on_session_roster()
{
     // We only want to receive the OnRoster signal /once/
     // more than that is just excessive
     _roster_conn.disconnect();

     // Now we're connected and ready to go
     evtConnected();
}

void JabberConnection::on_session_disconnected()
{
     // Disconnect the socket...
     _Transmitter->disconnect();
}

void JabberConnection::on_session_version(string& name, string& version, string& os)
{
     name    = ENV_VARS.package;
     version = ENV_VARS.version;
     // Run uname
     struct utsname osinfo;
     uname(&osinfo);
     os = string(osinfo.sysname) + " " + string(osinfo.machine);
}

void JabberConnection::on_session_time(string& UTF8Time, string& UTF8TimeZone)
{
     UTF8Time = Glib::locale_to_utf8(UTF8Time);
     UTF8TimeZone = Glib::locale_to_utf8(UTF8TimeZone);
}

} // namespace Gabber
