#ifndef ROSTERMODEL_HH
#define ROSTERMODEL_HH

#include "JabberConnection.hh"

#include <vector>
#include <list>

#include <gtkmm/treemodel.h>
#include <gdkmm/pixbuf.h>

#include <jabberoo.hh>

namespace Gabber
{

class RosterModel_Class;

/**
 * Model for a Gtk::TreeView that is based on the roster state
 */
class RosterModel : public Glib::Object, public Gtk::TreeModel
{
public:

    struct ModelColumns : public Gtk::TreeModelColumnRecord 
    {
        Gtk::TreeModelColumn<Glib::ustring> jid;
        Gtk::TreeModelColumn<Glib::ustring> name;
        Gtk::TreeModelColumn<Glib::RefPtr<Gdk::Pixbuf> > icon;
        Gtk::TreeModelColumn<bool> use_icon;
        Gtk::TreeModelColumn<bool> is_group;

        ModelColumns()
        { add(jid); add(name); add(icon); add(use_icon); add(is_group); }
    };
    /**
    * The list of available columns.
    */
    const ModelColumns columns;

    RosterModel(jabberoo::Session& sess, bool hide_offline = true, 
                bool sort_with_groups = true);
    ~RosterModel();

     /**
      * Clear the RosterModel.
      */
     void clear();

    enum SortType { rmSortAvailability, rmSortFullName };

    // XXX Not used yet
    void setSortType(SortType type);
    SortType getSortType() const
    { return _sort_type; }

    /// use groups in the sorting methods
    void setSortWithGroups(bool use_groups)
    { _sort_with_groups = use_groups; refresh(); }
    bool getSortWithGroups() const
    { return _sort_with_groups; }

    /// Hide users that are offline
    void setHideOffline(bool hide)
    { _hide_offline = hide; refresh(); }
    bool getHideOffline() const
    { return _hide_offline; }

protected:
    // Inheritted from Gtk::TreeModel
    void set_value_impl(const iterator& row, int column, const Glib::ValueBase& value);
    void get_value_impl(const iterator& row, int column, Glib::ValueBase& value) const;

    Gtk::TreeModelFlags get_flags_vfunc();
    int get_n_columns_vfunc();
    GType get_column_type_vfunc(int index);
    bool iter_next_vfunc(GtkTreeIter* iter);
    bool iter_children_vfunc(GtkTreeIter* iter, const GtkTreeIter* parent);
    bool iter_has_child_vfunc(const GtkTreeIter* iter);
    int iter_n_children_vfunc(const GtkTreeIter* iter);
    bool iter_nth_child_vfunc(GtkTreeIter* iter, const GtkTreeIter* parent, int n);
    bool iter_parent_vfunc(GtkTreeIter* iter, const GtkTreeIter* child);
    void ref_node_vfunc(GtkTreeIter* iter);
    void unref_node_vfunc(GtkTreeIter* iter);
    TreeModel::Path get_path_vfunc(const TreeModel::iterator& iter);
    bool get_iter_vfunc(GtkTreeIter* iter, const TreeModel::Path& path);
    void get_value_vfunc(const TreeModel::iterator& iter, int column, GValue* value);

    // Roster signals
    void on_roster_refresh(void);
    void on_roster_pres(const std::string&, bool available, 
            jabberoo::Presence::Type prev_type);

    // Roster XPath
    void on_roster_set(const judo::Element& elem);
private:
    friend class RosterModel_Class;
    static RosterModel_Class rostermodel_class_;

    struct sortCaseInsensitive :
        std::binary_function<const char*, const char*, bool>
    {
        bool operator()(const char* lhs, const char* rhs)
        { return strcasecmp(lhs, rhs) < 0; }
    };

    struct sortGroups :
        std::binary_function<const Glib::ustring&, const Glib::ustring&, bool>
    {
        bool operator()(const Glib::ustring& lhs, const Glib::ustring& rhs)
        {
            return (lhs.casefold().compare(rhs.casefold())) < 0;
        }
    };

    struct sortNickname :
        std::binary_function<const jabberoo::Roster::Item*, 
                             const jabberoo::Roster::Item*, bool>
    {
        bool operator()(const jabberoo::Roster::Item* lhs, 
                        const jabberoo::Roster::Item* rhs) const
        {
            return sortCaseInsensitive()(lhs->getNickname().c_str(),
                                         rhs->getNickname().c_str());
        }
    };

    /**
    * Houses extra information for a roster item.
    * This is primarily used to cache things that might be slow to lookup a lot.
    */
    struct ItemInfo
    {
        int icon_id;
    };

    typedef std::set<jabberoo::Roster::Item*, sortNickname> RosterItems;
    typedef std::map<Glib::ustring, RosterItems, sortGroups> ModelItems;
    typedef std::map<jabberoo::Roster::Item*, ItemInfo> RosterItemInfos;

    jabberoo::Session&                          _sess;
    jabberoo::Roster&                           _roster;
    ModelItems                                  _items;
    RosterItemInfos                             _item_info;
    bool                                        _hide_offline;
    SortType                                    _sort_type;
    bool                                        _sort_with_groups;
    bool                                        _has_first_refresh;
    XPath::Query*                               _roster_xpath;
    std::vector<Glib::RefPtr<Gdk::Pixbuf> >     _icons;

    void refresh();
    void addRosterItem(jabberoo::Roster::Item& item);
    void delRosterItem(jabberoo::Roster::Item& item);
    Gtk::TreeModel::Path getPathFromIter(const GtkTreeIter* iter);
    void clearIter(GtkTreeIter* iter);
    void clearRosterItems(const ModelItems::key_type& key);
    int getIconID(const std::string& jid);
    
}; // class RosterModel

class RosterModel_Class : public Glib::Class
{
public:
    struct RosterModelClass
    {
        GObjectClass parent_class;
    };

    friend class RosterModel;

    const Glib::Class& init();

    static void class_init_function(void* g_class, void* class_data);
}; // class RosterModel_Class

}; // namespace Gabber

#endif // ROSTERMODEL_HH
