#include "GabberApp.hh"
#include "GabberWidgets.hh"
#include "GabberUtility.hh"
#include "ChatView.hh"
#include "ChatViewManager.hh"
#include "LogManager.hh"

#include <sigc++/slot.h>

#include <gtkmm/messagedialog.h>
#include <gtkmm/scrolledwindow.h>
#include <gtkmm/stock.h>

namespace Gabber {

// -------------------------------------------------------------------
//
// Chat (One-on-One) Message View
//
// -------------------------------------------------------------------

ChatView::ChatView(ChatViewManager& mgr, const Element& msg)
     :BaseGabberWindow("OOOChat_win"), _mgr(mgr),
     _jid(msg.getAttrib("from")), _sent_composing(false), _composing_id(""),
     _last_msg_time(time(NULL))
{
     init(false);// Basic initialization
     on_event_node(msg);
     display(msg);
}

ChatView::ChatView(ChatViewManager& mgr, const std::string& jid)
     :BaseGabberWindow("OOOChat_win"), _mgr(mgr),
     _jid(jid), _sent_composing(false), _composing_id(""),
     _last_msg_time(time(NULL))
{
     init(true);// Basic initialization
}


ChatView::~ChatView()
{
    delete _txtChatview;
}

void ChatView::close()
{
    if(!_composing_id.empty() && _sent_composing)
    {
        // They deleted, stop the event
        Message m(_jid, "", Message::mtChat);
        Element* x = m.addX("jabber:x:event");
        x->addElement("id", _composing_id);

        G_App.getSession() << m;
        
        _sent_composing = false;
    }

    // Don't get xpaths anymore, that'd be bad
    G_App.getSession().unregisterXPath(_chat_query);
    G_App.getSession().unregisterXPath(_event_query);
    G_App.getSession().unregisterXPath(_presence_query);

    // unregister chat
    _mgr.releaseChat(_jid);

    // destroy the Gtk::Window
    BaseGabberWindow::close();
}
    
void ChatView::init(bool is_blank)
{
     // Get widgets
     get_widget("AddContact_btn", _btnAddContact);
     get_widget("SendAsMsg_tglbtn", _tglbtnSendAsMsg);
     get_widget("Message_txt", _txtMessage);

     // Figure out the line height
     // THEME/FONT SPECIFIC
     Glib::RefPtr<Pango::Layout> playout = Pango::Layout::create(_txtMessage->get_pango_context());
     Pango::Rectangle inkrect, logicalrect;
     playout->get_pixel_extents(inkrect, logicalrect);

     // Set the text entry area to be 2 lines tall per default
     _txtMessage->set_size_request(-1, logicalrect.get_height() * 2 + 4);

     Gtk::ScrolledWindow* scroll_win;
     get_widget("Chatview_scroll", scroll_win);
     _txtChatview = new PlainTextView(scroll_win);
     _txtChatview->append_timestamp();
     scroll_win->show_all();

//     manage(new GabberDnDText(_memMsg)); // Attach some drag and drop logic
//     _baseCloseChat = manage(new BaseGabberWidget("OOOChat_CloseChat_dlg", "OOOChat_win"));

     // Nickname and status display
     // if there's a resource
     if (is_blank && !JID::getResource(_jid).empty())
	  // We allow resource selection
	  _prettyjid = manage(new PrettyJID(_jid, "", PrettyJID::dtNickRes, true));
     else
	  _prettyjid = manage(new PrettyJID(_jid, "", PrettyJID::dtNickRes, false));
     _prettyjid->show();
     _prettyjid->changed.connect(slot(*this, &ChatView::on_resource_changed));

     Gtk::HBox* hb;
     get_widget("JIDInfo_hbox", hb);
     hb->pack_start(*_prettyjid, true, true, 0);
     _nickname = _prettyjid->get_nickname();

     // Set the nicknames
     _local_nick = G_App.getSession().getUserName();



//     _baseCloseChat->getLabel("OOOChat_CloseChat_lbl")->set_text(substitute(_("You just received a message from %s."), fromUTF8(_thisWindow, pj->get_nickname()))); 

     // Create chat view
//   _chatview = new ChatView(_thisWindow, _hboxChatview, false);

     // Connect events
     _btnAddContact->signal_clicked().connect(slot(*this, &ChatView::on_AddContact_clicked));
     _thisWindow->signal_event().connect(slot(*this,
                 &ChatView::on_window_event));

//     G_App->getSession().evtPresence.connect(slot(this, &ChatView::on_presence));
//     G_App->getErrorManager().errorChat.connect(slot(this, &ChatView::on_error));

//   Current implementation lacks such buttons
     Gtk::Button* b = 0;
//     getButton("OOOChat_Send_btn")->clicked.connect(slot(this, &ChatView::OnSend));
//     getButton("OOOChat_Close_btn")->clicked.connect(slot(this, &ChatView::on_Close_clicked));
     get_widget("ViewContact_btn", b);
     b->signal_clicked().connect(slot(*this, &ChatView::on_ViewContact_clicked));
     get_widget("History_btn", b);
     b->signal_clicked().connect(slot(*this, &ChatView::on_History_clicked));

//     _baseCloseChat->getButton("OOOChat_CloseChat_Keep_btn")->clicked.connect(slot(this, &ChatView::on_CloseChat_Keep_clicked));
//     _baseCloseChat->getButton("OOOChat_CloseChat_Close_btn")->clicked.connect(slot(this, &ChatView::on_CloseChat_Close_clicked));

     // Tweak widgets
//     _dlgCloseChat = static_cast<Gnome::Dialog*>(_baseCloseChat->get_this_widget());
//     _dlgCloseChat->set_close(false);
//     _dlgCloseChat->close_hides(true);
//     _dlgCloseChat->set_parent(*_thisWindow);

//      if (_onRoster)
// 	  _btnAddContact->set_sensitive(false);
//      else
// 	  _btnAddContact->set_sensitive(!G_App->isGroupChatID(_jid));

     // Set the window title
     _thisWindow->set_title(Util::substitute(_("Chat with %s"), _prettyjid->get_nickname()));

//     _thisWindow->set_default_size(G_App->getCfg().chats.ooowidth, G_App->getCfg().chats.oooheight);

    // put msgs in the text box

     // Register to recieve chat messages from _jid
     _chat_query = G_App.getSession().registerXPath(
             "/message[@type='chat' and @from='" + _jid + "']",
             slot(*this, &ChatView::on_chat_node));
     // Register to recieve chat events addressed from _jid
     _event_query = G_App.getSession().registerXPath(
             "/message[@from='" + _jid + 
             "']/x[@xmlns='jabber:x:event']", slot(*this, 
                 &ChatView::on_event_node));

     // Register for presence events
     _presence_query = G_App.getSession().registerXPath(
	  "/presence[@from='" + _jid + "']", 
	  slot(*this, &ChatView::on_presence_node));
		
     // Grab any existing presence
     try {
	  const Presence& p = G_App.getSession().presenceDB().findExact(_prettyjid->get_full_jid());
	  on_presence_node(p.getBaseElement());
     } catch (PresenceDB::XCP_InvalidJID& e) {
	  Presence p("", Presence::ptUnavailable, Presence::stOffline, _("No presence has been received."));
	  on_presence_node(p.getBaseElement());
     }
     
     
     // Display the window
     show();
     _thisWindow->raise();
}

void ChatView::on_Send_clicked()
{
    // if we're selecting a resource
    if (_prettyjid->is_selecting_resource())
    {
	    // Then lock down the resource
	    on_resource_locked();
    }

    // Add a timestamp if the gap has been too large
    // XXX configurable timestamp gap?
    time_t t = time(NULL);
    if ( (t - _last_msg_time) >= 300 )
    {
        _txtChatview->append_timestamp();
    }
    _last_msg_time = t;

    Glib::RefPtr<Gtk::TextBuffer> tb = _txtMessage->get_buffer();
    Glib::ustring body = tb->get_text(tb->begin(), tb->end());
    
    // Bogus
    if (body.empty())
        return;

    // Construct message
    Message m(_jid, body, Message::mtChat);

    // XXX Need disco or something here to check this a better way
    m.requestComposing();
    m.setID(G_App.getSession().getNextID());

    if (!_thread.empty())
	 m.setThread(_thread);

    G_App.getSession() << m; // Send the message
    LogManager::getSingleton().log(_jid) << m.toString() << std::endl;

    _txtChatview->cancel_composing("");
    _txtChatview->append(m, _local_nick, true);

    _txtMessage->get_buffer()->set_text(Glib::ustring()); // reset the message area
    _sent_composing = false; // don't send a cancel composing message, because we haven't sent composing
}

bool ChatView::on_delete_event(GdkEventAny* e)
{
    time_t now = time(NULL);
    // XXX configurable time somehow?
    if ( (now - _last_msg_time) <= 3 )
    {
        Gtk::MessageDialog dlg(*_thisWindow, 
            Util::substitute(_("A message was just received from %s, are you sure you want to close this chat?"), _nickname),
            Gtk::MESSAGE_WARNING, Gtk::BUTTONS_NONE, true);
	dlg.set_title(Util::substitute(_("Close chat with %s"), _nickname));
	dlg.add_button(_("_Don't Close"), Gtk::RESPONSE_NO);
	dlg.add_button(Gtk::Stock::CLOSE, Gtk::RESPONSE_YES);

        int ret = dlg.run();
        dlg.hide();
        if (ret == Gtk::RESPONSE_NO)
            return 1;
    }

    BaseGabberWindow::on_delete_event(e);
}

bool ChatView::on_window_event(GdkEvent* ev) 
{
    if (ev->type != GDK_KEY_PRESS)
    return false;

    GdkEventKey* e = (GdkEventKey*)ev;

    // escape closes our window
    if (e->keyval == GDK_Escape)
    {
        close();
        return false;
    }

    // If they pressed the Keypad enter, make it act like a normal enter
    if (e->keyval == GDK_KP_Enter)
        e->keyval = GDK_Return;

    if (e->keyval == GDK_Return)
    {
        //enter a newline if shift-return is used
        if (e->state & GDK_SHIFT_MASK)
        {
            //unset the shift bit. shift-return seems to have a special meaning for the widget
            e->state ^= GDK_SHIFT_MASK;
            return false;
        }

        // Ctrl-Return sends the message so we don't need to check for it  
        on_Send_clicked();
        return true;
    }

    // Normal typing, send the appropiate composing msg
    int chars = _txtMessage->get_buffer()->get_char_count();
    if (chars == 0)
    {
        if(!_composing_id.empty() && _sent_composing)
        {
            // They deleted, stop the event
            Message m(_jid, "", Message::mtChat);
            Element* x = m.addX("jabber:x:event");
            x->addElement("id", _composing_id);

            G_App.getSession() << m;
            
            _sent_composing = false;
        }

        _txtChatview->cancel_composing("");
        return false;
    }
    else if (chars > 0)
    {
        if (!_composing_id.empty() && !_sent_composing)
        {
            // Send a fresh composing
            Message m(_jid, "", Message::mtChat);
            Element* x = m.addX("jabber:x:event");
            x->addElement("composing");
            x->addElement("id", _composing_id);

            G_App.getSession() << m;

            _sent_composing = true;
        }

        _txtChatview->composing("", _local_nick);
    }

    return false;
}

void ChatView::on_AddContact_clicked()
{
     // Start AddContactDruid with that JID
//     AddContactDruid::display(JID::getUserHost(_jid));
}

void ChatView::on_ViewContact_clicked()
{
//      if (_onRoster)
//           ContactInfoDlg::display(_jid, _thisWindow);
//      else
//           ContactInfoDlg::display(_jid, Roster::rsNone, _thisWindow);
}

void ChatView::on_History_clicked()
{
//      string file = "file://";
//      file += string(G_App->getLogFile(_jid));
//      gnome_url_show(file.c_str());
}

void ChatView::on_chat_node(const Element& e)
{
    // XXX If this window is not visible we copy into _pending_msgs

    LogManager::getSingleton().log(_jid) << e.toString() << std::endl;
    display(Message(e));
}

void ChatView::on_event_node(const Element& e)
{
    std::cout << "Handling event: " << e.toString() << std::endl;
    Message m(e);

    const judo::Element* x = m.findX("jabber:x:event");
    if (!x)
        return;
    const judo::Element* delivered = x->findElement("delivered");
    const judo::Element* composing = x->findElement("composing");
    const judo::Element* id = x->findElement("id");

    if (composing && id)
    {
        std::cout << "Adding composing msg for " << e.getAttrib("from") <<
            std::endl;
        // If we have both of these they are typing
        _txtChatview->composing(e.getAttrib("from"), _nickname);
    }
    else if (!composing && id)
    {
        // cancel the events
        _txtChatview->cancel_composing(e.getAttrib("from"));
    }
    else if (composing && !id)
    {
        _composing_id = m.getID();
        std::cout << "Reset composing id to " << _composing_id << std::endl;
    }
    else if (delivered)
    {
        // Always handle these, if they are in there
        G_App.getSession() << m.delivered();
    }
}

void ChatView::on_presence_node(const Element& e)
{
     Presence p(e);
     
     Gtk::HBox* hb;
     get_widget("Show_hbox", hb);

     Gtk::Label* l;
     get_widget("Status_lbl", l);

     // Show the extra bar if they are not available or chatty
     if (p.getShow() == Presence::stOnline || p.getShow() == Presence::stChat)
     {
	  hb->hide();
     }
     else
     {
	  hb->show();
	  // If there is an explanation to go along with the short presence, display it properly
	  if (!p.getStatus().empty())
	  {
	       l->set_label("<b>" + Util::getShowName(p.getShow()) + "</b>: " + p.getStatus());
	  }
	  else
	  {
	       l->set_label("<b>" + Util::getShowName(p.getShow()) + "</b>");
	  }
     }
}

void ChatView::on_resource_changed()
{
     G_App.getSession().unregisterXPath(_presence_query);

     // Register for presence events
     _presence_query = G_App.getSession().registerXPath(
	  "/presence[@from='" + _jid + "']", 
	  slot(*this, &ChatView::on_presence_node));
		
     // Grab any existing presence
     try {
	  const Presence& p = G_App.getSession().presenceDB().findExact(_prettyjid->get_full_jid());
	  on_presence_node(p.getBaseElement());
     } catch (PresenceDB::XCP_InvalidJID& e) {
	  Presence p("", Presence::ptUnavailable, Presence::stOffline, _("No presence has been received."));
	  on_presence_node(p.getBaseElement());
     }
}

void ChatView::on_resource_locked()
{
     // hide the resource selection now
     _prettyjid->hide_resource_select();

     // unregister the general jid
     G_App.getSession().unregisterXPath(_chat_query);
     G_App.getSession().unregisterXPath(_event_query);
     _mgr.releaseChat(_jid);
     
     // get the new jid
     _jid = _prettyjid->get_full_jid();

     // if there's no resource, we probably want to do something
     // like catch whatever message is sent back.. ?
     if (JID::getResource(_jid).empty())
     {
	  cout << "empty resource selected" << endl;
     }
     
     // register the specific jid
     _mgr.register_chat(_jid, this);
     // Register to recieve chat messages from _jid
     _chat_query = G_App.getSession().registerXPath(
	  "/message[@type='chat' and @from='" + _jid + "']",
	  slot(*this, &ChatView::on_chat_node));
     // Register to recieve chat events addressed from _jid
     _event_query = G_App.getSession().registerXPath(
	  "/message[@from='" + _jid + 
	  "']/x[@xmlns='jabber:x:event']", slot(*this, 
						&ChatView::on_event_node));
}

void ChatView::display(const Message& m)
{
    // Track the last time we got a message to warn on fast close, we also put
    // in a timestamp if there has been a gap of 5 minutes or more
    time_t t = time(NULL);
    if ( (t - _last_msg_time) >= 300 )
    {
        _txtChatview->append_timestamp();
    }
    _last_msg_time = t;

    // bail on useless nodes
    if (m.getBody().empty()) return;

    // Display it
    _txtChatview->append(m, _nickname);

    // Save the thread!
    _thread = m.getThread();

    // Say we were displayed if they want that
    Element* x = m.findX("jabber:x:event");
    if (x && x->findElement("displayed"))
        G_App.getSession() << m.displayed();
}

} // namespace Gabber
